
var lpadAlign = require('lpad-align')
var longest = require('longest')
var chalk = require('chalk')
var tsml = require('tsml')

var isPlainObj = function (o) { return o !== null && typeof o === 'object' && o.constructor === Object; }

var isLiteral = function (val) { return typeof val === 'boolean' || val === null || val === undefined; }

var annotate = function (formatter, keyword, val) { return tsml(["\n  ", "\n  ", "\n  ", "\n"], formatter.annotation(("[" + keyword + " ")), formatter.string(val), formatter.annotation(']')); }

var formatFunction = function (formatter, functionType, fn) { return annotate(formatter, functionType, fn.displayName || fn.name || 'anonymous'); }

var formatCollapsedObject = function (formatter, val) { return tsml(["\n  ", "\n  ", "\n  ", "\n"], formatter.punctuation('('), formatter.string('collapsed'), formatter.punctuation(')')); }

var formatValue = function (formatter, val) {
  if (typeof val === 'number') {
    return formatter.number(val)
  }

  if (isLiteral(val)) {
    return formatter.literal(String(val))
  }

  var stringified = Object.prototype.toString.call(val)

  if (stringified === '[object Function]') {
    return formatFunction(formatter, 'Function', val)
  }

  if (stringified === '[object GeneratorFunction]') {
    return formatFunction(formatter, 'GeneratorFunction', val)
  }

  return tsml(["\n    ", "\n    ", "\n    ", "\n  "], formatter.punctuation('"'), formatter.string(val), formatter.punctuation('"'))
}

var isIterableWithKeys = function (val) { return isPlainObj(val) || Array.isArray(val); }

var formatWithDepth = function (obj, depth, formatter, offset) {
  var keys = Object.keys(obj)
  var coloredKeys = keys.map(function (key) { return formatter.property(key); })
  var colon = ': '

  var parts = keys.map(function (key, i) {
    var val = obj[key]

    var out = tsml(["\n      ", "\n      ", "\n    "], lpadAlign(coloredKeys[i], coloredKeys, offset), formatter.punctuation(colon))

    if (depth.curr > depth.max) {
      return out + formatCollapsedObject(formatter, val)
    }

    if (isIterableWithKeys(val)) {
      out += formatWithDepth(
        val,
        { curr: depth.curr + 1, max: depth.max },
        formatter,
        offset + longest(keys).length + colon.length
      )
    } else {
      out += formatValue(formatter, val)
    }

    return out
  })

  return '\n' + parts.join('\n')
}

var defaultFormatter = {
  punctuation: chalk.yellow,
  annotation: chalk.gray,
  property: chalk.green,
  literal: chalk.magenta,
  number: chalk.cyan,
  string: chalk.bold
}

var format = function (
  obj,
  depth,
  formatter,
  offset
) {
  if ( depth === void 0 ) depth = Infinity;
  if ( formatter === void 0 ) formatter = defaultFormatter;
  if ( offset === void 0 ) offset = 2;

  return formatWithDepth(obj, { curr: 0, max: depth }, formatter, offset);
}

module.exports = format

